// maxlisp v 0.8, Brad Garton (7/2004) (not really changed much from v 0.7, 4/2004)
// uses the clisp bundled executable lib
// see http://music.columbia.edu/cmc/maxlisp for more info
//
// invaluable assistance writing this max/msp object from Dan Trueman, R. Luke Dubois and Joshua Kit Clayton


#include "ext.h"
#include "z_dsp.h"
#include "string.h"
#include "ext_strings.h"
#include "edit.h"
#include "ext_wind.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define MAX_INPUTS 10 	//arbitrary
#define MAX_OUTPUTS 10	//also arbitrary
#define MAX_SCRIPTS 20	//how many scripts can we store internally


/******* maxlisp stuff *******/
typedef int (*maxlispmainFunctionPtr)(char *lispmemfileloc);
typedef void (*pumplispFunctionPtr)();
typedef void (*setlispbufFunctionPtr)(char *text, int nchars);
typedef char* (*getoutbufFunctionPtr)();


void *maxlisp_class;

typedef struct _maxlisp
{
	//header
    t_pxobject x_obj;
    
    //variables specific to this object
    long num_inputs, num_outputs; 	//number of inputs and outputs
    float in[MAX_INPUTS];			//values of input variables
    float in_connected[MAX_INPUTS]; //booleans: true if signals connected to the input in question
    								//we use this "connected" boolean so that users can connect *either* signals or floats
    								//to the various inputs; sometimes it's easier just to have floats, but other times
    								//it's essential to have signals.... but we have to know.
    void *outpointer;
    
    /******* maxlisp stuff *******/
	maxlispmainFunctionPtr maxlispmain;
	pumplispFunctionPtr pumplisp;
	setlispbufFunctionPtr setlispbuf;
	getoutbufFunctionPtr getoutbuf;

	CFBundleRef maxlispBundle;


	// for the maxlisp_var() and maxlisp_varlist() $n variable scheme
	#define NVARS 9	
	float var_array[NVARS];
	Boolean var_set[NVARS];
	
	int parenflag;
	int doublequotesflag;
	char retbuffer[8192]; // matches the 8192 in lisp.run.so, see "unixaux.c" -- totally stupid

	// editor stuff
    struct ed *c_edit;
    char *maxlisp_script[MAX_SCRIPTS], s_name[MAX_SCRIPTS][256];
    long maxlisp_script_len[MAX_SCRIPTS];
    short current_script, path[MAX_SCRIPTS];
 } t_maxlisp;

int n_maxlispes;  // for unloading if we reach the end

// for where the maxlisp folder is located
char *pathptr;
char pathname[1000];


/****PROTOTYPES****/

//setup funcs; this probably won't change, unless you decide to change the number of
//args that the user can input, in which case maxlisp_new will have to change
void *maxlisp_new(long num_inoutputs);
void maxlisp_assist(t_maxlisp *x, void *b, long m, long a, char *s);
void maxlisp_free(t_maxlisp *x);

//for getting floats at inputs
void maxlisp_float(t_maxlisp *x, double f);
void maxlisp_bang(t_maxlisp *x);

//for custom messages
void maxlisp_text(t_maxlisp *x, Symbol *s, short argc, Atom *argv);
void maxlisp_dotext(t_maxlisp *x, Symbol *s, short argc, Atom *argv);
void maxlisp_var(t_maxlisp *x, Symbol *s, short argc, Atom *argv);
void maxlisp_varlist(t_maxlisp *x, Symbol *s, short argc, Atom *argv);
void maxlisp_parens(t_maxlisp *x, Symbol *s, short argc, Atom *argv);
void maxlisp_doublequotes(t_maxlisp *x, Symbol *s, short argc, Atom *argv);


//for the text editor
void maxlisp_edclose (t_maxlisp *x, char **text, long size);
void maxlisp_dblclick(t_maxlisp *x);
void maxlisp_goscript(t_maxlisp *x, Symbol *s, short argc, Atom *argv);
void maxlisp_dogoscript(t_maxlisp *x, Symbol *s, short argc, Atom *argv);
void maxlisp_openscript(t_maxlisp *x, Symbol *s, short argc, Atom *argv);
void maxlisp_read(t_maxlisp *x, Symbol *s, short argc, Atom *argv);
void maxlisp_doread(t_maxlisp *x, Symbol *s, short argc, t_atom *argv);
void maxlisp_write(t_maxlisp *x, Symbol *s, short argc, Atom *argv);
void maxlisp_writeas(t_maxlisp *x, Symbol *s, short argc, Atom *argv);
void maxlisp_dowrite(t_maxlisp *x, Symbol *s, short argc, t_atom *argv);
void maxlisp_okclose (t_maxlisp *x, char *prompt, short *result); 

// do the lisp
char *maxlisp_golisp(t_maxlisp *x, char *lispbuf);

/****FUNCTIONS****/

//primary MSP funcs
void main(void)
{
	int i;
	short path, rval;
	
	//the A_DEFLONG argument gives us the object argument for the user to set number of ins/outs
	//change these if you want different user args
    setup((struct messlist **)&maxlisp_class, (method)maxlisp_new, (method)maxlisp_free, (short)sizeof(t_maxlisp), 0L, A_DEFLONG, 0);
   
	//standard messages; don't change these  
    addmess((method)maxlisp_assist,"assist", A_CANT,0);
    
    //our own messages
    addmess((method)maxlisp_text, "text", A_GIMME, 0);
    addmess((method)maxlisp_text, "maxlisp", A_GIMME, 0); // nothing special has to happen for "maxlisp"
    addmess((method)maxlisp_var, "var", A_GIMME, 0);
	addmess((method)maxlisp_varlist, "varlist", A_GIMME, 0);
	addmess((method)maxlisp_parens, "parens", A_GIMME, 0);
	addmess((method)maxlisp_doublequotes, "doublequotes", A_GIMME, 0);


    //so we know what to do with floats that we receive at the inputs
    addfloat((method)maxlisp_float);
    
    // and brad's bang...
    addbang((method)maxlisp_bang);

    //for the text editor
    addmess ((method)maxlisp_edclose, "edclose", A_CANT, 0); 
    addmess((method)maxlisp_dblclick,	"dblclick",	A_CANT, 0);
    addmess((method)maxlisp_goscript, "goscript", A_GIMME, 0);
    addmess((method)maxlisp_openscript, "openscript", A_GIMME, 0);
    addmess((method)maxlisp_read, "read", A_GIMME, 0);
	addmess ((method)maxlisp_okclose, "okclose", A_CANT, 0);  
    addmess((method)maxlisp_write, "savescript", A_GIMME, 0);
    addmess((method)maxlisp_writeas, "savescriptas", A_GIMME, 0);
    
    
    //gotta have this one....
    finder_addclass(0L, "maxlisp");
    
    n_maxlispes = 0;

 	// find the maxlisp folder location
 	path = 0;
 	rval = -1;
	nameinpath("maxlisp-bundle", &path);
	if (path != 0) // David Kiers suggested this, possibly a source of some crashes if the bundle isn't installed properly
		rval = path_topathname(path, "", pathname);
	if (rval != 0) error("couldn't find the maxlisp-bundles folder!");
	else {
		for (i = 0; i < 1000; i++)
			if (pathname[i] == '/') break;
		pathptr = pathname+i;
	}

//post("the pathname to the bundle: %s",pathname);

    // ho ho!
    post("maxlisp -- v 0.8");
}


//this gets called when the object is created; everytime the user types in new args, this will get called
void *maxlisp_new(long num_inoutputs)
{
	int i;	
    t_maxlisp *x;
    
    // maxlisp vars
	CFURLRef    maxlispbundleURL;
	UInt8 pathname[200];
	char lispinitmempath[1024];
	Boolean didLoad = false;

	    
    //leave this; creates our object
    x = (t_maxlisp *)newobject(maxlisp_class);

    //zero out the struct, to be careful (takk to jkclayton)
    if (x) { 
        for(i=sizeof(t_pxobject);i<sizeof(t_maxlisp);i++)  
                ((char *)x)[i]=0; 
	}
	
	// these are the entry function pointers in to the maxlisp.so lib
	x->maxlispmain = NULL;
	x->pumplisp = NULL;
	x->setlispbuf = NULL;
	x->getoutbuf = NULL;
	

	//constrain number of inputs and outputs
	//could also just define these internally, and not have user control via arguments....
	if (num_inoutputs < 1) num_inoutputs = 1;
	if (num_inoutputs > MAX_INPUTS) num_inoutputs = 1;
	
	//obviously, you could ignore the user args if you don't want them to control
	//the number of ins/outs, and just set them here.
	x->num_inputs = num_inoutputs;
	x->num_outputs = num_inoutputs;

    
    //outputs
    // right-to-left, bang first
    // x->bangout = bangout((t_object *)x);
    x->outpointer = outlet_new((t_object *)x, 0); // for bangs (from MAXBANG), values + value lists (from MAXMESSAGE)

    //initialize some variables; important to do this!
    for (i=0;i<x->num_inputs;i++){
    	x->in[i] = 0.;
    	x->in_connected[i] = 0;
    }
	
	//occasionally this line is necessary if you are doing weird asynchronous things with the in/out vectors
	//x->x_obj.z_misc = Z_NO_INPLACE;

	n_maxlispes++; // still keeping track so that we don't unload the executable while other maxlisp objects are around
	x->parenflag = 1;
	x->doublequotesflag = 1;

	// I'm reloading the bundle every time here, probably shouldn't.  I'm not sure it really does.
	// In rtcmix~ this is done with a kludgey sym-link to "dummy" bundles so that I get unique entry points
	// for each instance, but in maxlisp they all go to the same one.  This is ok for maxlisp, and in fact probably desired
    maxlispbundleURL = CFURLCreateWithFileSystemPath(
                    kCFAllocatorDefault,
    				CFStringCreateWithFormat(NULL, NULL, CFSTR("%s"), pathptr),
                    kCFURLPOSIXPathStyle,
                    true );


	// this error trap and the one later aren't working, fix 'em
	if (!maxlispbundleURL) {
		post("ERROR: maxlisp couldn't find the maxlisp.so library.  Sorry!");
	}
	
    // Make a bundle instance using the URLRef.
    x->maxlispBundle = CFBundleCreate( kCFAllocatorDefault, maxlispbundleURL );

	// I use this in an error message below
	CFURLGetFileSystemRepresentation(maxlispbundleURL, true, pathname, 200);

	// load it up!
	didLoad = CFBundleLoadExecutable(x->maxlispBundle);

	// find all the function entry points in the loaded executable lib
	if (didLoad)
	{
		x->maxlispmain = (void *)CFBundleGetFunctionPointerForName(x->maxlispBundle, CFSTR("maxlispmain"));

		if (x->maxlispmain)
		{
			sprintf(lispinitmempath, "%slispinit.mem", pathptr);
			x->maxlispmain(lispinitmempath);
		} else {
			error("maxlisp could not call maxlispmain()");
		}
		
		x->pumplisp = (void *)CFBundleGetFunctionPointerForName(x->maxlispBundle, CFSTR("pumplisp"));
		if (!(x->pumplisp)) {
			error("maxlisp could not find pumplisp()");
		}
		
		x->setlispbuf = (void *)CFBundleGetFunctionPointerForName(x->maxlispBundle, CFSTR("setlispbuf"));
		if (!(x->setlispbuf)) {
			error("maxlisp could not find setlispbuf()");
		}

		x->getoutbuf = (void *)CFBundleGetFunctionPointerForName(x->maxlispBundle, CFSTR("getoutbuf"));
		if (!(x->getoutbuf)) {
			error("maxlisp could not find getoutbuf()");
		}
	} else {
		error("maxlisp could not find the maxlisp.so library");
		error("  here's where I looked: %s",pathname);
	}

	// set up for the variable-substitution scheme	
	for(i = 0; i < NVARS; i++) {
		x->var_set[i] = false;
		x->var_array[i] = 0.0;
	}

    //the text editor
    x->c_edit = 0;
    x->current_script = 0;
    for (i = 0; i < MAX_SCRIPTS; i++) {
    	x->s_name[i][0] = 0;
    }

    return (x);
}




//tells the user about the inputs/outputs when mousing over them
void maxlisp_assist(t_maxlisp *x, void *b, long m, long a, char *s)
{

	if (m == 1) {
		sprintf(s, "lisp in");
	}
	if (m == 2) {
		sprintf(s, "something out");
	}
}

// here's my free function
void maxlisp_free(t_maxlisp *x)
{
	// keep track of how many and which maxlispN.so libs have been loaded, unload when last one done
	if (--n_maxlispes == 0)
		CFBundleUnloadExecutable(x->maxlispBundle);
}


//this gets called when ever a float is received at *any* input
// unused in maxlisp, but maybe in The Future...
void maxlisp_float(t_maxlisp *x, double f)
{
	int i;
	
	//check to see which input the float came in, then set the appropriate variable value
	for(i=0;i<x->num_inputs;i++) {
		if (x->x_obj.z_in == i) {
			x->in[i] = f;
			post("maxlisp: setting in[%d] =  %f, but maxlisp doesn't use this?!?", i, f);
		} 
	}
}


// here's brad banging around
void maxlisp_bang(t_maxlisp *x)
{
	post("maxlisp, v. 0.8 by Brad Garton");
}

char* maxlisp_golisp(t_maxlisp *x, char *lispbuf)
{
	int i, blen;
	int parabal, npumps;
	char *obufptr;

	parabal = 0;
	npumps = 0;
	blen = strlen(lispbuf);
	
	for (i = 0; i < blen; i++) {
		if ( (lispbuf[i] != ' ') && (lispbuf[i] != '\n') ) { // get rid of leading spaces and CRs
			if (lispbuf[i] == '(') parabal++; // pop on left parens
			else if (lispbuf[i] == ')') parabal--; // pop off right parens
			if (parabal == 0) npumps++; // they have balanced
		}
	}
	
	if (parabal != 0) return("ERROR: the parentheses aren't balanced!");
	
	x->setlispbuf(lispbuf, blen);

	for (i = 0; i < npumps; i++) x->pumplisp();
	
	obufptr = (x->getoutbuf())+4; // this is to get rid if the godawful prompt nonsense in clisp
	blen = strlen(obufptr);
	for (i = 0; i < blen; i++) {
		if (obufptr[i] == 10) x->retbuffer[i] = ' ';
		else x->retbuffer[i] = obufptr[i];
	}
	x->retbuffer[blen] = '\0';

	return(x->retbuffer);
}

// see note for maxlisp_dotext() below
void maxlisp_text(t_maxlisp *x, Symbol *s, short argc, Atom *argv)
{
	defer_low(x, (method)maxlisp_dotext, s, argc, argv); // always defer this message
}

// what to do when we get the message "text" or the message "maxlisp"
// maxlisp scores come from the [textedit] or message-box objects this way
void maxlisp_dotext(t_maxlisp *x, Symbol *s, short argc, Atom *argv)
{
	short i, varnum;
#define INBUFSIZE 8192
	char thebuf[INBUFSIZE];
	char xfer[1024];
	char *bptr;
	int nchars;
	int top;
	char *obptr;

	
	for (i = 0; i < INBUFSIZE; i++) thebuf[i] = 0;  // gotta do this or thebuf retains stuff
	bptr = thebuf;
	nchars = 0;
	top = 0;
	
	for (i=0; i < argc; i++) {
		switch (argv[i].a_type) {
			case A_LONG:
				sprintf(xfer, " %ld", argv[i].a_w.w_long);
				break;
			case A_FLOAT:
				sprintf(xfer, " %lf", argv[i].a_w.w_float);
				break;
			case A_DOLLAR:
				varnum = argv[i].a_w.w_long;
				if ( !(x->var_set[varnum-1]) ) {
					error("variable $%d has not been set yet, LISP not called", varnum);
					return;
				}
				sprintf(xfer, " %f", x->var_array[varnum-1]);
				break;
			case A_SYM:
				if (top == 0) { sprintf(xfer, "%s", argv[i].a_w.w_sym->s_name); top = 1;}
				else sprintf(xfer, " %s", argv[i].a_w.w_sym->s_name);
				break;
			case A_SEMI:
				sprintf(xfer, ";");
				break;
			case A_COMMA:
				sprintf(xfer, ",");
		}
		strcpy(bptr, xfer);
		nchars = strlen(xfer);
		bptr += nchars;
	}

//post("got a buffer: %s	%d", thebuf,strlen(thebuf));
	if (strlen(thebuf) == 0) return; // guard against sending 0-length buffers

	obptr = maxlisp_golisp(x, thebuf);
	
	if (x->parenflag == 0)
		for (i = 0; i < strlen(obptr); i++)
			if ( (obptr[i] == '(') || (obptr[i] == ')') ) obptr[i] = ' ';

	if (x->doublequotesflag == 0)
		for (i = 0; i < strlen(obptr); i++)
			if (obptr[i] == '"') obptr[i] = ' ';

	outlet_anything(x->outpointer, gensym(obptr), 0, NIL);
}


// the "var" message allows us to set $n variables imbedded in a scorefile with varnum value messages
void maxlisp_var(t_maxlisp *x, Symbol *s, short argc, Atom *argv)
{
	short i, varnum;

	for (i = 0; i < argc; i += 2) {
		varnum = argv[i].a_w.w_long;
		if ( (varnum < 1) || (varnum > NVARS) ) {
			error("only vars $1 - $9 are allowed");
			return;
		}
		x->var_set[varnum-1] = true;
		switch (argv[i+1].a_type) {
			case A_LONG:
				x->var_array[varnum-1] = (float)(argv[i+1].a_w.w_long);
				break;
			case A_FLOAT:
				x->var_array[varnum-1] = argv[i+1].a_w.w_float;
		}
	}
}


// the "varlist" message allows us to set $n variables imbedded in a scorefile with a list of positional vars
void maxlisp_varlist(t_maxlisp *x, Symbol *s, short argc, Atom *argv)
{
	short i;

	if (argc >= NVARS) {
			error("asking for too many variables, only setting the first 10 ($0-$9)");
			argc = NVARS;
	}
	
	for (i = 0; i < argc; i++) {
		x->var_set[i] = true;
		switch (argv[i].a_type) {
			case A_LONG:
				x->var_array[i] = (float)(argv[i].a_w.w_long);
				break;
			case A_FLOAT:
				x->var_array[i] = argv[i].a_w.w_float;
		}
	}
}

// the "parens" message turns on/off the printing of lisp parens on output
// [parens 0] == off;  [parens 1] == 1
void maxlisp_parens(t_maxlisp *x, Symbol *s, short argc, Atom *argv)
{
	if (argc != 1) post("ERROR: the maxlisp [parens] message needs a 0 or 1");
	switch (argv[0].a_type) {
		case A_LONG:
			x->parenflag = argv[0].a_w.w_long;
	}
	if ( !((x->parenflag == 0) || (x->parenflag == 1)) ) post("ERROR: the maxlisp [parens] message needs a 0 or 1");
}

// the "doublequotes" message turns on/off the printing of lisp string double-quotes(") on output
// [doublequotes 0] == off;  [doublequotes 1] == 1
void maxlisp_doublequotes(t_maxlisp *x, Symbol *s, short argc, Atom *argv)
{
	if (argc != 1) post("ERROR: the maxlisp [doublequotes] message needs a 0 or 1");
	switch (argv[0].a_type) {
		case A_LONG:
			x->doublequotesflag = argv[0].a_w.w_long;
	}
	if ( !((x->doublequotesflag == 0) || (x->doublequotesflag == 1)) ) post("ERROR: the maxlisp [doublequotes] message needs a 0 or 1");
}


// here is the text-editor buffer stuff, go dan trueman go!
// used for maxlisp internal buffers
void maxlisp_edclose (t_maxlisp *x, char **text, long size) 
{
	if (x->maxlisp_script[x->current_script]) {
		t_freebytes(x->maxlisp_script[x->current_script], x->maxlisp_script_len[x->current_script] * sizeof(char));
		x->maxlisp_script[x->current_script] = 0;
	}
	x->maxlisp_script_len[x->current_script] = size;
    x->maxlisp_script[x->current_script] = t_getbytes(size * sizeof(char));
	strncpy(x->maxlisp_script[x->current_script], *text, size);
	x->c_edit = 0;
}

void maxlisp_okclose (t_maxlisp *x, char *prompt, short *result)
{
	*result = 3; //don't put up dialog box
	return;
}

void maxlisp_dblclick(t_maxlisp *x) 
{ 
	char title[80];
	
	if (x->c_edit) {
		if(x->maxlisp_script[x->current_script]) 
			ed_settext(x->c_edit, &x->maxlisp_script[x->current_script], x->maxlisp_script_len[x->current_script]);
		ed_vis(x->c_edit); 
	}
	else {
		x->c_edit = ed_new((Object *)x); 
		sprintf(title,"script_%d", x->current_script);
		wind_settitle(x->c_edit->e_wind, title, 0);
		if(x->maxlisp_script[x->current_script]) 
			ed_settext(x->c_edit, &x->maxlisp_script[x->current_script], x->maxlisp_script_len[x->current_script]);
	} 
}

// see note for maxlisp_dogoscript() below
void maxlisp_goscript(t_maxlisp *x, Symbol *s, short argc, Atom *argv)
{
	defer_low(x, (method)maxlisp_dogoscript, s, argc, argv); // always defer this message  
}


// the [goscript N] message will cause buffer N to be sent to the maxlisp parser
void maxlisp_dogoscript(t_maxlisp *x, Symbol *s, short argc, Atom *argv)
{
	short i,j,temp;
	int tval;
	int buflen;
	char thebuf[8192]; // should probably by dyn-alloced, or at least set to coordinate with maxlisp (of necessary)
	char *obptr;

	if (argc == 0) {
		error("goscript needs a buffer number [0-19]");
		return;
	}
		
	for (i = 0; i < argc; i++) {
		switch (argv[i].a_type) {
			case A_LONG:
				temp = (short)argv[i].a_w.w_long;
				if (temp > MAX_SCRIPTS) temp = MAX_SCRIPTS;
				if (temp < 0) temp = 0;
				break;
			case A_FLOAT:
				temp = (short)argv[i].a_w.w_float;
    			if (temp > MAX_SCRIPTS) temp = MAX_SCRIPTS;
				if (temp < 0) temp = 0;
				break;
		}
	}
	x->current_script = temp;

	buflen = x->maxlisp_script_len[x->current_script];

	// probably don't need to transfer to a new buffer, but I want to be sure there's room for the \0,
	// plus the substitution of \n for those annoying ^M thingies	
	for (i = 0, j = 0; i < buflen; i++) {
		thebuf[j] = *(x->maxlisp_script[x->current_script]+i);
		if ((int)thebuf[j] == 13) thebuf[j] = '\n'; // maxlisp wants newlines, not <cr>'s
		
		// ok, here's where we substitute the $vars
		if (thebuf[j] == '$') {
			sscanf(x->maxlisp_script[x->current_script]+i+1, "%d", &tval);
			if ( !(x->var_set[tval-1]) ) {
				error("variable $%d has not been set yet, LISP not called", tval);
				return;
			}
			sprintf(thebuf+j, "%f", x->var_array[tval-1]);
			j = strlen(thebuf)-1;
			i++; // skip over the var number in input
		}
		j++;
	}
	thebuf[j] = '\0';

	obptr = maxlisp_golisp(x, thebuf);

	outlet_anything(x->outpointer, gensym(obptr), 0, NIL);
}

// [openscript N] will open a buffer N
void maxlisp_openscript(t_maxlisp *x, Symbol *s, short argc, Atom *argv)
{
	short i,temp;
	
	for (i = 0; i < argc; i++) {
		switch (argv[i].a_type) {
			case A_LONG:
				temp = (short)argv[i].a_w.w_long;
				if (temp > MAX_SCRIPTS) temp = MAX_SCRIPTS;
				if (temp < 0) temp = 0;
				break;
			case A_FLOAT:
				temp = (short)argv[i].a_w.w_float;
    			if (temp > MAX_SCRIPTS) temp = MAX_SCRIPTS;
				if (temp < 0) temp = 0;
				break;
		}
	}
	x->current_script = temp;
	maxlisp_dblclick(x);
}


void maxlisp_write(t_maxlisp *x, Symbol *s, short argc, Atom *argv)  
{    
	short err, i, temp;

	for (i = 0; i < argc; i++) {
		switch (argv[i].a_type) {
			case A_LONG:
				temp = (short)argv[i].a_w.w_long;
				if (temp > MAX_SCRIPTS) temp = MAX_SCRIPTS;
				if (temp < 0) temp = 0;
				x->current_script = temp;
				post("maxlisp: current script is %d", temp);
				break;
			case A_FLOAT:
				temp = (short)argv[i].a_w.w_float;
    			if (temp > MAX_SCRIPTS) temp = MAX_SCRIPTS;
				if (temp < 0) temp = 0;
				x->current_script = temp;
				post("maxlisp: current script is %d", temp);
				break;
		}
	}
	
	defer(x,(method)maxlisp_dowrite,s,argc,argv); // always defer this message  
}

void maxlisp_writeas(t_maxlisp *x, Symbol *s, short argc, Atom *argv)  
{   
	short err, i, temp;
	
	for (i=0; i < argc; i++) {
		switch (argv[i].a_type) {
			case A_LONG:
				temp = (short)argv[i].a_w.w_long;
				if (temp > MAX_SCRIPTS) temp = MAX_SCRIPTS;
				if (temp < 0) temp = 0;
				x->current_script = temp;
				x->s_name[x->current_script][0] = 0;
				post("maxlisp: current script is %d", temp);
				break;
			case A_FLOAT:
				temp = (short)argv[i].a_w.w_float;
    			if (temp > MAX_SCRIPTS) temp = MAX_SCRIPTS;
				if (temp < 0) temp = 0;
				x->current_script = temp;
				x->s_name[x->current_script][0] = 0;
				post("maxlisp: current script is %d", temp);
				break;
			case A_SYM://this doesn't work yet
				strcpy(x->s_name[x->current_script], argv[i].a_w.w_sym->s_name);
				post("maxlisp: writing file %s",x->s_name[x->current_script]);
				break;
		}
	}
	
	defer(x,(method)maxlisp_dowrite,s,argc,argv); // always defer this message  
}

void maxlisp_dowrite(t_maxlisp *x, Symbol *s, short argc, t_atom *argv)
{
	char filename[256];
	t_handle script_handle;
	short err, i, temp;
	short type = 'TEXT'; 
	long type_chosen, thistype = 'TEXT';
	long size;
	short outtype, count;
	t_filehandle fh;
	
	if(!x->s_name[x->current_script][0]) {
		//if (saveas_dialog(&x->s_name[0][x->current_script], &x->path[x->current_script], &type)) 
		  if (saveasdialog_extended(x->s_name[x->current_script], &x->path[x->current_script], &type_chosen, &thistype, 1))
			return; //user cancelled
	} 
	strcpy(filename, x->s_name[x->current_script]);
	
	err = path_createsysfile(filename, x->path[x->current_script], thistype, &fh);  
	if (err) {       
		fh = 0;      
		error("maxlisp: error %d creating file", err); 
		return;
	}
	
	script_handle = sysmem_newhandle(0);
	sysmem_ptrandhand (x->maxlisp_script[x->current_script], script_handle, x->maxlisp_script_len[x->current_script]); 

	err = sysfile_writetextfile(fh, script_handle, TEXT_LB_UNIX); 
	if (err) {       
		fh = 0;      
		error("maxlisp: error %d writing file", err); 
		return;
	}
	
	sysmem_freehandle(script_handle);
	sysfile_seteof(fh, x->maxlisp_script_len[x->current_script]); 
	sysfile_close(fh);

	return;	

}

void maxlisp_read(t_maxlisp *x, Symbol *s, short argc, Atom *argv)  
{    
	defer(x,(method)maxlisp_doread,s,argc,argv); // always defer this message  
}

void maxlisp_doread(t_maxlisp *x, Symbol *s, short argc, t_atom *argv) 
{
	char filename[256];
	short err, i, temp;
	long type = 'TEXT'; 
	long size;
	long outtype, count;
	t_filehandle fh;
	t_handle script_handle;
	
	for (i = 0; i < argc; i++) {
		switch (argv[i].a_type) {
			case A_LONG:
				temp = (short)argv[i].a_w.w_long;
				if (temp > MAX_SCRIPTS) temp = MAX_SCRIPTS;
				if (temp < 0) temp = 0;
				x->current_script = temp;
				x->s_name[x->current_script][0] = 0;
				break;
			case A_FLOAT:
				temp = (short)argv[i].a_w.w_float;
    			if (temp > MAX_SCRIPTS) temp = MAX_SCRIPTS;
				if (temp < 0) temp = 0;
				x->current_script = temp;
				x->s_name[x->current_script][0] = 0;
				break;
			case A_SYM:
				strcpy(filename, argv[i].a_w.w_sym->s_name);
				strcpy(x->s_name[x->current_script], filename);
		}
	}
	
	if(!x->s_name[x->current_script][0]) {
//		if (open_dialog(filename, &path, &outtype, &type, 1))
		if (open_dialog(filename,  &x->path[x->current_script], &outtype, 0L, 0)) // allow all types of files

			return; //user cancelled
	} else {
		if (locatefile_extended(filename, &x->path[x->current_script], &outtype, &type, 1)) {
			error("maxlisp: error opening file: can't find file"); 
			x->s_name[x->current_script][0] = 0;
			return; //not found
		}
	}
	
	//we should have a valid filename at this point
	err = path_opensysfile(filename, x->path[x->current_script], &fh, READ_PERM);  
	if (err) {       
		fh = 0;      
		error("error %d opening file", err); 
		return;
	}
	
	strcpy(x->s_name[x->current_script], filename);
	
	sysfile_geteof(fh, &size);
	if (x->maxlisp_script[x->current_script]) {
		t_freebytes(x->maxlisp_script[x->current_script], x->maxlisp_script_len[x->current_script] * sizeof(char));
		x->maxlisp_script[x->current_script] = 0;
	}
	if (!(x->maxlisp_script[x->current_script] = t_getbytes(size)) || !(script_handle = sysmem_newhandle(size))) {
		error("maxlisp: %s too big to read", filename);
		return;
	} else {
		x->maxlisp_script_len[x->current_script] = size;
		sysfile_readtextfile(fh, script_handle, size, TEXT_LB_NATIVE);
		strcpy(x->maxlisp_script[x->current_script], *script_handle);
	}
	sysmem_freehandle(script_handle);
	sysfile_close(fh);

	return;
	
}

